{{
        Code from the ajv Terminal emulator
''
'' Based on code from Vince Briel and ajv

}}
OBJ
    text:       "VGA_1024"              ' VGA Terminal Driver

VAR
    'terminal emulation vars
    byte force7 '  force7 - Flag force to 7 bits
    byte autolf '  autolf - Generate LF after CR
    byte state  ' Main terminal emulation state
    long a0     '  Arg 0 to an escape sequence
    long a1     '   ...arg 1
    byte onlast ' Flag that we've just put a char on last column
    word pos    ' Current output/cursor position

    byte savemins ' # minutes until blank screen
    byte color  ' Text color (index)
    byte baud   ' Baud rate (index)
    byte cursor ' Cursor style
    word regTop, regBot ' Scroll region top/bottom
    byte lastc  ' Last displayed character

PUB start(video)
    ' Start VGA output driver
    text.start(video)

'' Initialize terminal emulator code
PUB init

    text.cls

    ' Init state vars
    state := 0
    onlast := 0
    pos := 0
    regTop := 0
    regBot := text#chars

PUB prn2(val) | dig
    dig := 48 + (val // 10)
    val := val/10
    if val > 0
        prn2(val)
    text.putc(pos++, dig)

PUB prn(val)
    text.putc(pos++, " ")
    if val < 0
        text.putc(pos++, "-")
        val := 0 - val
    prn2(val)

PUB putn(r, c, val)
    pos := r * text#cols + c
    prn(val)

'' Write a string into the screen
PUB puts(row, col, str) | x, ptr
    ptr := (row * text#cols) + col
    repeat x from 0 to STRSIZE(str)-1
        text.putc(ptr++, BYTE[str+x])


'' Process next byte from our host port
PUB singleSerial0(c)
    case state

    ' State 0: ready for new data to display or start of escape sequence
     0:
        ' Assume high bit chars are alternate character set
        '  output, and make them consume a space
        if c > 127
            c := $20

        ' Printing chars; put on screen
        if c => 32
            simplec(c)
            text.setCursorPos(pos)
            return

        ' Escape sequence started
        if c == 27
            state := 1
            return

        ' CR
        if c == 13
            pos := pos - (pos // text#cols)
            text.setCursorPos(pos)
            return

        ' LF
        if c == 10
            if inReg
                pos += text#cols
                if pos => regBot
                    scrollUp
                    pos -= text#cols
            else
                pos += text#cols
                if pos => text#chars
                    pos -= text#cols
            text.setCursorPos(pos)
            return

        ' Tab
        if c == 9
            ' Advance to next tab stop
            pos += (8 - (pos // 8))

            ' Scroll when tab to new line
            if pos => text#chars
                pos := text#lastline
                text.delLine(0)
            text.setCursorPos(pos) 
            return

        ' Backspace
        if c == 8
            if pos > 0
                pos -= 1
            text.setCursorPos(pos) 
            return

    ' State 1: ESC received, ready for escape sequence
     1:
        case c

         ' ESC-[, start of extended ANSI style arguments
         "[":
            a0 := a1 := -1
            state := 2
            return

         ' ESC-P, cursor down one line
         "P":
            pos += text#cols
            if pos => text#chars
                pos -= text#cols

         ' ESC-K, cursor left one position
         "K":
            if pos > 0
                pos -= 1

         ' ESC-H, cursor up one line
         "H":
            pos -= text#cols
            if pos < 0
                pos += text#cols

         ' ESC-D, scroll one line
         "D":
            if inReg
                scrollUp

         ' ESC-M, scroll backward
         "M":
            if inReg
                scrollDown

         ' ESC-G, cursor home
         "G":
            pos := 0

         ' ESC-(, char set selection (decoded and ignored)
         "(":
            state := 5
            return

        ' Escape sequence done, reset state machine
        state := 0
        text.setCursorPos(pos) 
        return

    ' State 2: ESC-[, start decoding first numeric arg
     2:
        ' Digits, assemble value
        if (c => "0") AND (c =< "9")
            if a0 == -1
                a0 := c - "0"
            else
                a0 := (a0*10) + (c - "0")
            return

        ' Semicolon, advance to arg1
        if c == ";"
            state := 3
            return

        ' End of input sequence
        ansi(c)
        text.setCursorPos(pos) 
        return

    ' State 3: ESC-[<digits>;, start decoding second numeric arg
     3:
        ' Digits, assemble value
        if (c => "0") AND (c =< "9")
            if a1 == -1
                a1 := c - "0"
            else
                a1 := (a1*10) + (c - "0")
            return

        ' Semicolon, ignore subsequent args
        if c == ";"
            state := 4
            return

        ' End of sequence
        ansi(c)
        text.setCursorPos(pos) 
        return

    ' State 4: ESC-[<digits>;<digits>;...  Ignore subsequent args
     4:
        if (c => "0") AND (c =< "9")
            return
        if c == ";"
            return
        ansi(c)
        text.setCursorPos(pos) 
        return

    ' State 5: ESC-(, ignore character set selection
     5:
        state := 0
        text.setCursorPos(pos) 
        return
    return

'' Set invert video based on control code
PRI setInv(c)
    if c == 1
        text.setInv(1)
    else
        text.setInv(0)

'' Tell if current position is within scroll region
PRI inReg : answer
    answer := (pos => regTop) AND (pos < regBot)

'' Scroll the contents of the scroll region upward
''  (i.e., add a new blank line at the bottom of the region)
PRI scrollUp
    text.delLine(regTop)
    if regBot < text#chars
        text.insLine(regBot)

'' Scroll downward (new blank line at top)
PRI scrollDown
    if regBot < text#chars
        text.delLine(regBot)
    text.insLine(regTop)

'' Take action for ANSI-style sequence
PRI ansi(c) | x, defVal

    ' Always reset input state machine at end of sequence
    state := 0

    ' Map args to appropriate default
    ' Most get a default argument of 1, a few 0.
    if (c <> "r") AND (c <> "J") AND (c <> "m") AND (c <> "K")
        if a0 == -1
            a0 := 1
        if a1 == -1
            a1 := 1

    case c

     "@":       ' Insert char(s)
        repeat while a0-- > 0
            text.insChar(pos)

     "b":       ' Repeat last char
        repeat while a0-- > 0
            simplec(lastc)

     "d":       ' Vertical position absolute
        if (a0 < 1) OR (a0 > text#rows)
            a0 := text#rows
        pos := ((a0-1) * text#cols) + (pos // text#cols)

     "m":       ' Set character enhancements
        setInv(a0)
        if a1 <> -1
            setInv(a1)

     "r":       ' Set scroll region
        ' TBD is to change all the scroll code to check the region

        ' Bound param to screen geometry
        if a0 < 1
            a0 := 1
        elseif a0 > text#cols
            a0 := text#cols
        if a1 < 1
            a1 := 1
        elseif a1 > text#cols
            a1 := text#cols
        if a1 < a0
            a1 := a0

        ' Set region; regTop is first location in the scroll region;
        '  regBot is first location beyond end of scroll region.
        regTop := (a0-1) * text#cols
        regBot := a1 * text#cols

        ' This op seems to implicitly home the cursor...
        pos := 0

     "A":       ' Move cursor up line(s)
        repeat while a0-- > 0
            pos -= text#cols
            if pos < 0
                pos += text#cols
                return

     "B":       ' Move cursor down line(s)
        repeat while a0-- > 0
            pos += text#cols
            if pos => text#chars
                pos -= text#cols
                return

     "C":       ' Move cursor right
        repeat while a0-- > 0
            pos += 1
            if pos => text#chars
                pos -= 1
                return

     "D":       ' Move cursor left
        repeat while a0-- > 0
            pos -= 1
            if pos < 0
                pos := 0
                return

     "G":       ' Horizontal position absolute
        if (a0 < 1) OR (a0 > text#cols)
            a0 := text#cols
        pos := (pos - (pos // text#cols)) + (a0-1)

     "H":       ' Set cursor position
        if a0 =< 0
            a0 := 1
        if a1 =< 0
            a1 := 1
        pos := (text#cols * (a0-1)) + (a1 - 1)
        if pos < 0
            pos := 0
        if pos => text#chars
            pos := text#chars-1

     "J":       ' Clear screen/EOS
        ' Erase to top of screen
        if a0 == 1
            text.clBOL(pos)
            x := pos - text#cols
            x -= x // text#cols
            repeat while x => 0
                text.clEOL(x)
                x -= text#cols
            return

        ' Erase whole screen and home cursor
        if a0 == 2
            pos := 0

        ' Clear from current position to end of screen
        text.clEOL(pos)
        x := pos + text#cols
        x -= (x // text#cols)
        repeat while x < text#chars
            text.clEOL(x)
            x += text#cols

     "K":       ' Clear parts of line
        if a0 == -1             ' No arg, to end of line
            text.clEOL(pos)
        elseif a0 == 1          ' 1 == from beginning to position
            text.clBOL(pos)
        else                    ' 2 == clear whole line
            text.clEOL(pos - (pos // text#cols))


     "L":       ' Insert line(s)
        if inReg
            repeat while a0-- > 0
                if regBot < text#chars
                    text.delLine(regBot)
                text.insLine(pos)

     "M":       ' Delete line(s)
        if inReg
            repeat while a0-- > 0
                text.delLine(pos)
                if regBot < text#chars
                    text.insLine(regBot)

     "P":       ' Delete char(s)
        repeat while a0--
            text.delChar(pos)

'' Put a single printing char onto the screen
PRI simplec(c)
    text.putc(pos++, lastc := c)

    ' If just advanced beyond end of scroll region, scroll
    if pos == regBot
        scrollUp
        pos -= text#cols

    ' If beyond scroll region & walk off screen, wrap
    '  back around to column 0
    elseif pos == text#chars
        pos := text#lastline
